;********************************************************************
;*                                                                  *
;* Title       : DCF                                                *
;* Version     : 0.3   25. September 2002                           *
;* Author      : Uwe Nagel, Parkstr. 46, D-68766 Hockenheim         *
;* Assembler   : MPASM (c) Microchip                                *
;*                                                                  *
;********************************************************************
; 0.1    21.05.02 erster Entwurf
; 0.2    30.05.02 erste lauffhige Version mit LCD
; 0.3    29.09.02 Kosmetik am Quellcode und an der Homepage, keine
;                 nderung der Funktion
; 0.4    01.04.03 Antennen Symbol auf Wunsch von Christof Rue
;********************************************************************

#define TITEL "-DCF77-Uhr V0.4-"

	title		TITEL
	__config	_XT_OSC & _WDT_ON & _PWRTE_ON & _CP_OFF
	errorlevel	-302		; keine bank select message

	include	<P16F84.INC>
	include "pic16cxy.inc"

;****** Variablen ***************************************************
 cblock	0x0c
	save_w		; w und
	save_status	; STATUS whrend interrupt
	flag		; diverse Bits
	temp
	tmp1
	tmp2
	zweihund	; 200 Interupts sind 1 Sek
	puls_zeit	; misst Pulsdauer
;*** diese Zeit wird angezeigt
	dispSec, dispMin, dispStd, dispTag, dispMon, dispJahr, dispWtag, dispStat
;*** hier kommt die gerade empfangene Zeit rein
	dcfMin, dcfStd, dcfTag, dcfMon, dcfJahr, dcfWtag, dcfStat
;*** zum Vergleich, die letzte Minute
	dcf1Min, dcf1Std, dcf1Tag, dcf1Mon, dcf1Jahr, dcf1Wtag
 endc


;****** bits in flag ************************************************
#define	neusync flag,0
#define	alt     flag,1
#define	sekunde	flag,2
#define	syncd	flag,3
#define	comm	flag,7

lc_data      equ tmp1       ; local in lcd_write
lcd_init_cnt equ tmp2       ; only used in lcd_init
temp_lcd     equ tmp2       ; temporary in write_lcd_char, clear_display
temp_byte    equ tmp2       ; temporary in write_byte


;****** Konstanten **************************************************
#define	TICKS	200	; fr 3,2768 MHz Quarz
;#define	TICKS	218	; fr NTSC-Quarz (hatte gerade nicht den richtigen)
; mit einem NTSC-Quarz luft die Uhr ungenau, wenn sie kein Signal empfngt

;********************************************************************
;* Programmcode ...
;********************************************************************
	org	0
	goto	main

;********************************************************************
;* interrupt :
;********************************************************************
	org	4
interrupt
	movwf	save_W		; W und STATUS sichern
	swapf	STATUS,w
	movwf	save_STATUS

	bcf	STATUS,RP0	; Whle register page 0

	jnb	INTCON,T0IF,end_t0_isr ; ist es timer interrupt ?
; bsf PortA,1       

	jb	alt,NoPos
	jnb	PortA,0,NoPos

;** positive Flanke...
	bsf	alt
; bsf PortA,2       
	clrf	puls_zeit

	jnb	neusync,NoNeg

	clrf	dispSec
	movf	dcfMin,w
	movwf	dispMin
	movf	dcfStd,w
	movwf	dispStd
	movf	dcfTag,w
	movwf	dispTag
	movf	dcfWtag,w
	movwf	dispWtag
	movf	dcfMon,w
	movwf	dispMon
	movf	dcfJahr,w
	movwf	dispJahr
	movf	dcfStat,w
	movwf	dispStat

	movlw	TICKS	; 218 fr NTSC Quartz sonst 200
	movwf	zweihund
	bcf	neusync
	bsf	sekunde
	goto	end_t0_isr
	
NoPos:	jnb	alt,NoNeg
	jb	PortA,0,NoNeg

;** fallende Flanke
	bcf	alt
; bcf PortA,2

	movf	puls_zeit,w
	addlw	-30
	call	shiftDCF

NoNeg:	decfsz	zweihund,f
	goto	TimeOut

;** Sekunde abgelaufen
	movlw	TICKS
	movwf	zweihund
; bsf PortA,3
	call	tick
; bcf PortA,3       
	bsf	sekunde


TimeOut:
	incfsz	puls_zeit,f	; 1,28 s keine steigende Flanke ?
	goto	end_t0_isr

	call	sec59


end_t0_isr
; bcf PortA,1       
	bcf	INTCON,T0IF     ; lsche timer interrupt flag
	swapf	save_status,w
	movwf	STATUS
	swapf	save_w,f
	swapf	save_w,w
	retfie
 
;****** Tabellen ****************************************************
lcd_init_table
	addwf	PCL,f
	retlw	0x02 | 0x80	; switch to 4 bit mode
	retlw	0x28 | 0x80	; better do it twice
	retlw	0x28 | 0x80	; init for 2x16 Display
	retlw	0x0c | 0x80	; Display on, Cursor off, no blinking
	retlw	0x06 | 0x80	; increment cursor, no display shift
	retlw	0x01 | 0x80	; Clear Display, Cursor home
	dt	TITEL		; Einschaltmeldung

	retlw	0x40 | 0x80	; Pointer auf CG-Ram

	retlw	0x00
	retlw	0x00
	retlw	0x00
	retlw	0x03
	retlw	0x04
	retlw	0x00
	retlw	0x01
	retlw	0x00

	retlw	0x00
	retlw	0x00
	retlw	0x00
	retlw	0x18
	retlw	0x04
	retlw	0x02
	retlw	0x11
	retlw	0x09

	retlw	0x06
	retlw	0x01
	retlw	0x05
	retlw	0x04
	retlw	0x0e
	retlw	0x0e
	retlw	0x1f
	retlw	0x1f

	retlw	0x09
	retlw	0x01
	retlw	0x02
	retlw	0x00
	retlw	0x00
	retlw	0x00
	retlw	0x00
	retlw	0x00

	retlw	0xff		; Ende der Tabelle

wtag_table
	addwf	PCL,f
	dt	"--MoDiMiDoFrSaSo"

;****** Hauptprogramm ***********************************************
main    

	bsf	STATUS,rp0	; select Register-Page 1
                  
	movlf	0x03,OPTION_REG	; TMR0 internal clock/16 = 200Hz @ 3,2768MHz Quarz
	movlf	0x01,TRISA	; porta4-porta1 output
	movlf	0x00,TRISB	; portb7..0 output
  
	bcf	STATUS,rp0	; select Register-Page 0

	clrf	flag
	clrf	puls_zeit
	movlf	TICKS,zweihund

	clrf	dispSec
	clrf	dispMin
	clrf	dispStd
	clrf	dispWtag
	clrf	dispJahr
	movlw	1
	movwf	dispMon
	movwf	dispTag

	movlf	0xa0,INTCON	; Enable  TMR0-Int


;------ Display initialisieren und Einschaltmeldung ------
	call	lcd_init	; init LC-Display
hallo:	clrwdt
	btfss	sekunde
	goto	hallo

	movlw	0x01		; Clear Display
	call	write_lcd_comm
;------

mainloop:
	clrwdt	
	btfss	sekunde
	goto	mainloop
	bcf	sekunde
	call	showtime
	goto	mainloop

;****** neue Sekunde anzeigen ***************************************
;  01234567889abcdef
; +-----------------+
; |20:03:15 MESZ ## |
; |Di 01.04.2003 ## |
; +-----------------+
showtime:
	movlw	0x80
	call	write_lcd_comm
	movf	dispStd,w
	call	write_byte
	movlw	':'
	call	write_lcd_data
	movf	dispMin,w
	call	write_byte
	movlw	':'
	call	write_lcd_data
	movf	dispSec,w
	call	write_byte
	movlw	' '
	call	write_lcd_data
	movlw	'M'
	call	write_lcd_data
	movlw	'E'
	call	write_lcd_data
	movlw	'S'
	btfsc	dispStat,4
	call	write_lcd_data
	movlw	'Z'
	call	write_lcd_data
	movlw	' '
	btfss	dispStat,4
	call	write_lcd_data
	movlw	' '
	call	write_lcd_data

	btfsc	syncd
	movlw	0x00
	btfss	syncd
	movlw	' '
	call	write_lcd_data
	btfsc	syncd
	movlw	0x01
	btfss	syncd
	movlw	' '
	call	write_lcd_data
	

	movlw	0xc0
	call	write_lcd_comm
	movf	dispWtag,w
	call	write_wtag
	movlw	' '
	call	write_lcd_data
	movf	dispTag,w
	call	write_byte
	movlw	'.'
	call	write_lcd_data
	movf	dispMon,w
	call	write_byte
	movlw	'.'
	call	write_lcd_data
	movlw	0x20
	call	write_byte
	movf	dispJahr,w
	call	write_byte
	movlw	' '
	call	write_lcd_data

	btfsc	syncd
	movlw	0x02
	btfss	syncd
	movlw	' '
	call	write_lcd_data
	btfsc	syncd
	movlw	0x03
	btfss	syncd
	movlw	' '
	call	write_lcd_data
	return
;********************************************************************

shiftDCF:
	rrf	dcfJahr,f
	rrf	dcfMon,f
	rrf	dcfTag,f
	rrf	dcfStd,f
	rrf	dcfMin,f
	rrf	dcfStat,f
	return

;********************************************************************
;55555555 54444444 44433333 33333222 22222221 11111111
;87654321 09876543 21098765 43210987 65432109 87654321   Sekunde
;PJJJJJJJ JMMMMMWW WTTTTTTP SSSSSSPm mmmmmm1a ooaR0000   DCF-Bits
;
;JJJJJJJJ 000MMMMM 00tttttt 00ssssss 0mmmmmmm 1aooaR00 00000www
; Jahr     Monat    Tag      Stunde   Minute   Status   Wtag
;********************************************************************

; jetzt werden die empfangenen Bits erstmal umsortiert 
; 
sec59:	rlf	dcfMon,w
	rlf	dcfJahr,f	; =Jahr
	rlf	dcfTag,w
	rlf	dcfMon,w
	andlw	0x07
	movwf	dcfWtag		; =Wochentag Mo..So = 1..7
	rrf	dcfMon,f
	rrf	dcfMon,f
	movlw	0x1f
	andwf	dcfMon,f	; =Monat
	rrf	dcfTag,f
	movlw	0x3f
	andwf	dcfTag,f	; =Tag
	rrf	dcfStd,w
	rrf	dcfMin,f
	rrf	dcfStat,f
	rrf	dcfMin,f
	rrf	dcfStat,f	; =Status
	movlw	0x7f
	andwf	dcfMin,f	; =Minute
	rrf	dcfStd,f
	rrf	dcfStd,f
	movlw	0x3f
	andwf	dcfStd,f	; =Stunde

	movlw	dcf1Min
	movwf	FSR
	call	tickMin		; letzte Zeit weiterrechnen

	movf	dcfMin,w	; und mit gerade empfangener
	xorwf	dcf1Min,w	; Zeit vergleichen
	skpz
	goto	ungleich
	movf	dcfStd,w
	xorwf	dcf1Std,w
	skpz
	goto	ungleich
	movf	dcfTag,w
	xorwf	dcf1Tag,w
	skpz
	goto	ungleich
	movf	dcfMon,w
	xorwf	dcf1Mon,w
	skpz
	goto	ungleich
	movf	dcfJahr,w
	xorwf	dcf1Jahr,w
	skpz
	goto	ungleich

;*** zwei aufeinanderfolgende Zeiten haben zueinander gepasst
;*** die aktuelle Zeit bei nchster Sekunde ins Display kopieren

	bsf	syncd
	bsf	neusync
	return

ungleich:
;*** zwei aufeinanderfolgende Zeiten haben nicht eine Minute unterschied
;*** merken wir die gerade empfangene Zeit fr die nchste Minute
	movf	dcfMin,w
	movwf	dcf1Min
	movf	dcfStd,w
	movwf	dcf1Std
	movf	dcfTag,w
	movwf	dcf1Tag
	movf	dcfMon,w
	movwf	dcf1Mon
	movf	dcfJahr,w
	movwf	dcf1Jahr
	bcf	syncd
	return

;------ hier wird die Zeit weitergezhlt
tick:	movlw	dispSec
	movwf	FSR

	call	IncDec	; Sekunde
	movlw	0x60
	xorwf	INDF,w
	skpz
	return
	clrf	INDF	; Sekunde=0

	incf	FSR,f	; ->Minute
tickMin:
	call	IncDec	; Minute
	movlw	0x60
	xorwf	INDF,w
	skpz
	return
	clrf	INDF	; Minute=0

	incf	FSR,f	; -> Stunde
	call	IncDec	; Stunde
	movlw	0x24
	xorwf	INDF,w
	skpz
	return
	clrf	INDF	; Stunde=0

	incf	FSR,f
tickTag:
	call	IncDec	; Tag

	incf	FSR,f	; ->Monat
	movlw	8
	subwf	INDF,w
	movf	INDF,w
	skpnc
	addlw	-7
	andlw	0x01	
	addlw	0x31
	movwf	temp
	movf	INDF,w
	xorlw	2	; Februar ?
	skpz
	goto	nofeb
	movlw	0x30
	movwf	temp
	incf	FSR,f	; ->Jahr
	movf	INDF,w	; Jahr
	decf	FSR,f	; ->Monat
	andlw	0x13
	skpnz
	goto	nofeb
	xorlw	0x12
	skpnz
	goto	nofeb
	movlw	0x29
	movwf	temp
nofeb:	decf	FSR,f	; -> Tag
	movf	temp,w
	xorwf	INDF,w
	skpz
	return
	movlw	1
	movwf	INDF	; Tag=1

	incf	FSR,f	; -> Monat
	call	IncDec	
	movf	INDF,w
	xorlw	0x13
	skpz
	return
	movlw	1
	movwf	INDF	; Monat=1

	incf	FSR,f	; -> Jahr
	call	IncDec	
	return

IncDec:	incf	INDF,f
	movf	INDF,w
	andlw	0x0f
	xorlw	0x0a
	skpz
	return
	movlw	0x06
	addwf	INDF,f
	return


;************* LCD-Routinen *****************************************
;********************************************************************
;* Display is in 4 bit mode. Datalines connected to PortB4..7
;* PortB3 = Register Select   PortB2 = Read not Write   PortB1 = Enable
;********************************************************************
LCEN equ 1
LCRW equ 2
LCRS equ 3

tst_lcd_busy
	movlf	(1<<LCRW),PortB
waitbusy
	bsf	PortB,LCEN		; get hi-Nibble
	rlf	PortB,w		; Busy-flag to Carry
	bcf	PortB,LCEN
	bsf	PortB,LCEN		; get lo-Nibble
	bcf	PortB,LCEN
	bc	waitbusy  
	return

;********************************************************************
write_lcd_comm
	bsf	comm
write_lcd_data
	movwf	lc_data
	call	tst_lcd_busy
	bsf	STATUS,RP0	; select Register-Page 1
	movlf	0x00,TRISB
	bcf	STATUS,RP0	; select Register-Page 0
	movf	lc_data,w
	call	lcd_nibble
	swapf	lc_data,w
	call	lcd_nibble

	bcf	comm
	bsf	STATUS,RP0	; select Register-Page 1
	movlf	0xf0,TRISB
	bcf	STATUS,RP0	; select Register-Page 0
	return

lcd_nibble  
	andlw	0xf0
	btfss	comm
	iorlw	(1<<LCRS)
	movwf	PortB
	bsf	PortB,LCEN
	bcf	PortB,LCEN
	return
;********************************************************************
lcd_init
	clrf	lcd_init_cnt
lcd_init_loop
	movf	lcd_init_cnt,w
	call	lcd_init_table
	xorlw	0xff		; mit 0xff vergleichen
	skpnz
	return
	xorlw	0xff		; xor rckgngig gemacht
	movwf	lc_data
	bsf	comm
	btfss	lc_data,7
	bcf	comm
	andlw	0x7f
	call	write_lcd_data
	incf	lcd_init_cnt,f
	goto	lcd_init_loop	
;********************************************************************
write_byte               ; writes a hex-byte
	movwf	temp_byte
	swapf	temp_byte,w
	call	write_nibble
	movf	temp_byte,w

write_nibble  
	andlw	0x0f
	addlw	0x06
	btfsc	STATUS,DC
	addlw	0x07
	addlw	0x30-0x06
	goto	write_lcd_data
;********************************************************************
write_wtag
	andlw	0x07
	movwf	temp_byte
	addwf	temp_byte,f
	movf	temp_byte,w
	call	wtag_table
	call	write_lcd_data
	incf	temp_byte,w
	call	wtag_table
	goto	write_lcd_data
;********************************************************************
	end	
